import fs from "node:fs";
import path from "node:path";
import { IncomingMessage, ServerResponse } from "node:http";

export interface Request extends Omit<IncomingMessage, 'url'> {
  url: {
    pathname: string;
    split: string[];
    searchParams: URLSearchParams;
    qs: Record<string, string>;
  };
  cookies?: Record<string, string>;
  params?: Record<string, string>;
  post?: Record<string, string>;
}

export interface Response extends ServerResponse {
  reply: (options: { code?: number; type?: string; body: string }) => void;
  status: (code: number) => Response;
  json: (body: JSON, code?: number) => void;
  html: (body: string, code?: number) => void;
  redirect: (target: string, code?: number) => void;
}

export type Handler = (req: Request, res: Response, next?: () => void) => void | Promise<void>;

export default class Router {
  private routes: Array<{
    path: string | RegExp;
    methods: { [method: string]: Handler[] }
  }> = [];
  private mimes: Map<string, string>;

  constructor() {
    this.mimes = new Map();
  }

  async importRoutesFromPath(p: string): Promise<this> {
    const dirEntries = await fs.promises.readdir(path.resolve() + '/' + p, { withFileTypes: true });
    for(const tmp of dirEntries) {
      if(tmp.isFile() && (tmp.name.endsWith(".mjs") || tmp.name.endsWith(".js"))) {
        const routeModule = (await import(`${path.resolve()}/${p}/${tmp.name}`)).default;
        this.use(routeModule(this));
      }
      else if(tmp.isDirectory()) {
        await this.importRoutesFromPath(p + '/' + tmp.name);
      }
    }
    return this;
  }

  group(basePath: string | RegExp, callback: (methods: any) => void): this {
    const self = this;

    const methods = {
      get(path: string | RegExp, ...handlers: Handler[]) {
        const fullPath = self.combinePaths(basePath, path);
        return self.registerRoute(fullPath, "get", handlers);
      },
      post(path: string | RegExp, ...handlers: Handler[]) {
        const fullPath = self.combinePaths(basePath, path);
        return self.registerRoute(fullPath, "post", handlers);
      },
      put(path: string | RegExp, ...handlers: Handler[]) {
        const fullPath = self.combinePaths(basePath, path);
        return self.registerRoute(fullPath, "put", handlers);
      },
      delete(path: string | RegExp, ...handlers: Handler[]) {
        const fullPath = self.combinePaths(basePath, path);
        return self.registerRoute(fullPath, "delete", handlers);
      },
      patch(path: string | RegExp, ...handlers: Handler[]) {
        const fullPath = self.combinePaths(basePath, path);
        return self.registerRoute(fullPath, "patch", handlers);
      }
    };
  
    callback(methods);

    return this;
  }

  private combinePaths(basePath: string | RegExp, subPath: string | RegExp): string | RegExp {
    if(typeof basePath === "string" && typeof subPath === "string")
      return `${basePath.replace(/\/$/, "")}/${subPath.replace(/^\//, "")}`;
    if(basePath instanceof RegExp && typeof subPath === "string")
      return new RegExp(`${basePath.source}${subPath.replace(/^\//, "")}`);
    if(typeof basePath === "string" && subPath instanceof RegExp)
      return new RegExp(`${basePath.replace(/\/$/, "")}${subPath.source}`);
    if(basePath instanceof RegExp && subPath instanceof RegExp)
      return new RegExp(`${basePath.source}${subPath.source}`);

    throw new TypeError("Invalid path types. Both basePath and subPath must be either string or RegExp.");
  }

  use(obj: Router): void {
    if(obj instanceof Router) {
      if(!Array.isArray(obj.routes))
        throw new TypeError("Routes must be an array.");
      this.routes = [ ...this.routes, ...obj.routes ];
      this.sortRoutes();
    }
  }

  get(path: string | RegExp, ...callback: Handler[]): this {
    this.registerRoute(path, "get", callback);
    return this;
  }

  post(path: string | RegExp, ...callback: Handler[]): this {
    this.registerRoute(path, "post", callback);
    return this;
  }

  head(path: string | RegExp, ...callback: Handler[]): this {
    this.registerRoute(path, "head", callback);
    return this;
  }

  put(path: string | RegExp, ...callback: Handler[]): this {
    this.registerRoute(path, "put", callback);
    return this;
  }

  delete(path: string | RegExp, ...callback: Handler[]): this {
    this.registerRoute(path, "delete", callback);
    return this;
  }

  patch(path: string | RegExp, ...callback: Handler[]): this {
    this.registerRoute(path, "patch", callback);
    return this;
  }

  private registerRoute(
    path: string | RegExp,
    method: string,
    handler: Handler[]
  ): this {
    const route = this.routes.find(route =>
      typeof route.path === "string"
        ? route.path === path
        : route.path instanceof RegExp && path instanceof RegExp && route.path.toString() === path.toString()
    );
  
    if(route) {
      route.methods[method] = [
        ...(route.methods[method] || []),
        ...handler
      ];
    }
    else {
      this.routes.push({
        path,
        methods: { [method]: handler }
      });
    }

    console.log("route set:", method.toUpperCase(), path);
    this.sortRoutes();
    return this;
  }

  getRoute(path: string, method: string): any {
    return this.routes
      .find(r => typeof r.path === "string"
        ? r.path === path
        : r.path.exec?.(path)
      && r.methods[method.toLowerCase()]
    );
  }

  private sortRoutes(): this {
    this.routes.sort((a, b) => {
      const aLength = typeof a.path === "string" ? a.path.length : a.path.source.length;
      const bLength = typeof b.path === "string" ? b.path.length : b.path.source.length;
  
      if(typeof a.path === "string" && typeof b.path === "string")
        return bLength - aLength;
      if(typeof a.path === "string")
        return -1;
      if(typeof b.path === "string")
        return 1;
      return bLength - aLength;
    });
    return this;
  }

  private readMimes(file: string = "/etc/mime.types"): void {
    fs.readFileSync(file, "utf-8")
      .split("\n")
      .filter(line => !line.startsWith("#") && line)
      .forEach(line => {
        const [mimeType, extensions] = line.split(/\s+/);
        extensions?.split(" ").forEach(ext => this.mimes.set(ext, mimeType));
      });
  }

  static({
    dir = path.resolve() + "/public",
    route = /^\/public/
  }: { dir?: string; route?: RegExp }): this {
    if(!this.mimes.size)
      this.readMimes();

    this.get(route, (req: Request, res: Response, next?: () => void) => {
      try {
        const filename = req.url.pathname.replace(route, "") || "index.html";
        const mime = this.mimes.get(filename.split(".").pop() || "");
        const file = path.join(dir, filename);
        const stat = fs.statSync(file);

        if(req.headers.range) {
          const [startStr, endStr] = req.headers.range.replace(/bytes=/, "").split("-");
          const start = parseInt(startStr, 10);
          const end = endStr ? parseInt(endStr, 10) : stat.size - 1;

          res.writeHead(206, {
            "Content-Range": `bytes ${start}-${end}/${stat.size}`,
            "Accept-Ranges": "bytes",
            "Content-Length": end - start + 1,
            "Content-Type": mime,
          });
          fs.createReadStream(file, { start, end }).pipe(res);
        }
        else {
          res.writeHead(200, {
            "Content-Length": stat.size,
            "Content-Type": mime,
          });
          fs.createReadStream(file).pipe(res);
        }
      }
      catch(err: any) {
        console.error(err);
        res.reply({ code: 404, body: "404 - File not found" });
      }
    });

    return this;
  }
}
